/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves
 *      shawn@talula.demon.co.uk
 *      http://www.talula.demon.co.uk/allegro/
 *
 *      3DNow! and SSE enhanced matrix routines
 *
 *      This file written by Calin Andrian.
 *
 *      See readme.txt for copyright information.
 */

#include "asmdefs.inc"
#include "mmxdefs.inc"
#include "mmx.h"

#ifdef ALLEGRO_MMX

.data

                 .align 4
__float_one:     .float 1.0, 0.0



.text



#define MTX_IN          ARG1
#define MTX_OUT         ARG2

.globl _build_matrix_al
   .align 4
_build_matrix_al:
   pushl %ebp
   movl %esp, %ebp
   pushl %ebx
   pushl %esi
   pushl %edi

   movl MTX_OUT, %edi
   movl MTX_IN, %esi
   addl $15, %edi
   andl $-16, %edi

   movl M_V00(%esi), %eax
   movl M_V01(%esi), %ebx
   movl M_V02(%esi), %ecx
   movl M_T0(%esi), %edx
   movl %eax, (%edi)
   movl %ebx, 4(%edi)
   movl %ecx, 8(%edi)
   movl %edx, 12(%edi)

   movl M_V10(%esi), %eax
   movl M_V11(%esi), %ebx
   movl M_V12(%esi), %ecx
   movl M_T1(%esi), %edx
   movl %eax, 16(%edi)
   movl %ebx, 20(%edi)
   movl %ecx, 24(%edi)
   movl %edx, 28(%edi)

   movl M_V20(%esi), %eax
   movl M_V21(%esi), %ebx
   movl M_V22(%esi), %ecx
   movl M_T2(%esi), %edx
   movl %eax, 32(%edi)
   movl %ebx, 36(%edi)
   movl %ecx, 40(%edi)
   movl %edx, 44(%edi)

   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret

#undef MTX_IN
#undef MTX_OUT



#define MTX     ARG1
#define X       ARG2
#define Y       ARG3
#define Z       ARG4
#define XOUT    ARG5
#define YOUT    ARG6
#define ZOUT    ARG7

.globl _apply_matrix_3dnow
   .align 4
_apply_matrix_3dnow:
   pushl %ebp
   movl %esp, %ebp

   FEMMS_R()

   movl MTX, %edx
   movd Z, %mm1
   addl $15, %edx
   movq X, %mm0
   andl $-16, %edx
   punpckldq __float_one, %mm1

   movq (%edx), %mm2
   movq 8(%edx), %mm3

   movq 16(%edx), %mm4
   PFMUL_R(0, 2)
   movq 24(%edx), %mm5
   PFMUL_R(1, 3)

   movq 32(%edx), %mm6
   PFMUL_R(0, 4)
   movq 40(%edx), %mm7
   PFMUL_R(1, 5)

   PFADD_R(2, 3)
   PFMUL_R(0, 6)
   PFADD_R(4, 5)
   PFMUL_R(1, 7)

   movl XOUT, %eax
   PFACC_R(5, 3)
   movl YOUT, %ecx
   PFADD_R(6, 7)
   movl ZOUT, %edx
   movd %mm3, (%eax)

   psrlq $32, %mm3
   movd %mm3, (%ecx)
   PFACC_R(7, 7)
   movd %mm7, (%edx)

   FEMMS_R()

   popl %ebp
   ret



#if 0 /* not yet */
.globl _apply_matrix_sse
   .align 4
_apply_matrix_sse:
   pushl %ebp
   movl %esp, %ebp

   movl MTX, %edx
   movl $1.0, XOUT
   addl $15, %edx
   MOVUPS_LD_D(5, 12, 0)        // movups 12(%ebp), %xmm0
   andl $-16, %edx

   MOVAPS_LD_B(2, 1)            // movaps (%edx), %xmm1
   MOVAPS_LD_D(2, 16, 2)        // movaps 16(%edx), %xmm2
   MOVAPS_LD_D(2, 32, 3)        // movaps 32(%edx), %xmm3

   MULPS_R(0, 1)                // mulps %xmm0, %xmm1
   MULPS_R(0, 2)                // mulps %xmm0, %xmm2
   MULPS_R(0, 3)                // mulps %xmm0, %xmm3

   MOVAPS_R(1, 4)               // movaps %xmm1, %xmm4
   MOVLHPS_R(2, 1)              // movlhps %xmm2, %xmm1
   MOVHLPS_R(4, 2)              // movhlps %xmm4, %xmm2
   MOVHLPS_R(3, 0)              // movhlps %xmm3, %xmm0

   ADDPS_R(2, 1)                // addps %xmm2, %xmm1
   ADDPS_R(3, 0)                // addps %xmm3, %xmm0

   MOVAPS_R(1, 2)               // movaps %xmm1, %xmm2
   SHUFPS_R(0, 1, 0x08)         // shufps $0x08, %xmm0, %xmm1
   movl XOUT, %eax
   SHUFPS_R(0, 2, 0x5d)         // shufps $0x5d, %xmm0, %xmm2
   movl YOUT, %ecx
   ADDPS_R(1, 2)                // addps %xmm1, %xmm2
   movl ZOUT, %edx

   MOVSS_ST_B(2, 0)             // movss %xmm2, (%eax)
   SHUFPS_R(2, 2, 0xa1)         // shufps $0xa1, %xmm2, %xmm2
   MOVSS_ST_B(2, 1)             // movss %xmm2, (%ecx)
   MOVHLPS_R(2, 2)              // movhlps %xmm2, %xmm2
   MOVSS_ST_B(2, 2)             // movss %xmm2, (%edx)

   popl %ebp
   ret
#endif /* not yet */

#undef MTX
#undef X
#undef Y
#undef Z
#undef XOUT
#undef YOUT
#undef ZOUT



#define X       ARG1
#define Y       ARG2
#define Z       ARG3
#define XOUT    ARG4
#define YOUT    ARG5

.globl _persp_project_3dnow
   .align 4
_persp_project_3dnow:
   pushl %ebp
   movl %esp, %ebp

   FEMMS_R()

   movd Z, %mm1
   movq X, %mm2

   movq __persp_xscale_f, %mm3
   PFRCP_R(1, 0)
   movq __persp_xoffset_f, %mm4
   PFMUL_R(3, 2)

   movl XOUT, %eax
   PFMUL_R(2, 0)
   movl YOUT, %edx
   PFADD_R(4, 0)

   movd %mm0, (%eax)
   psrlq $32, %mm0
   movd %mm0, (%edx)

   FEMMS_R()

   popl %ebp
   ret

#undef X
#undef Y
#undef Z
#undef XOUT
#undef YOUT



#define X       ARG1
#define Y       ARG2
#define Z       ARG3

.globl _normalize_vector_3dnow
   .align 4
_normalize_vector_3dnow:
   pushl %ebp
   movl %esp, %ebp

   FEMMS_R()

   movl X, %eax
   movl Z, %edx
   movl Y, %ecx

   movd (%eax), %mm0
   movd (%edx), %mm1
   punpckldq (%ecx), %mm0

   movq %mm1, %mm3
   PFMUL_R(1, 1)
   movq %mm0, %mm2
   PFMUL_R(0, 0)

   PFADD_R(1, 0)
   PFACC_R(0, 0)

   PFRSQRT_R(0, 1)
   movq %mm1, %mm7
   PFMUL_R(1, 1)
   PFRSQIT1_R(0, 1)
   PFRCPIT2_R(7, 1)

   PFMUL_R(1, 2)
   movd %mm2, (%eax)
   PFMUL_R(1, 3)
   movd %mm3, (%edx)
   psrlq $32, %mm2
   movd %mm2, (%ecx)

   FEMMS_R()

   popl %ebp
   ret

#undef X
#undef Y
#undef Z


#endif /* MMX */
