
#include <stdlib.h>
#include <stdio.h>

#include "allegro.h"
#include "p3d.h"

MATRIX mxi;
MATRIX_f mxf;
MATRIX_al mxal;
volatile fixed xi, yi, zi;
fixed xio, yio, zio;
volatile float xf, yf, zf;
float xfo, yfo, zfo;


int main()
{
   int i, j;

   allegro_init();
   check_cpu();

   for (i=0; i<3; i++) {
      for (j=0; j<3; j++) {
         mxf.v[i][j] = 3*i + 5*j + 13;
         mxi.v[i][j] = itofix(3*i + 5*j + 13);
      }
      mxf.t[i] = 7*i + 11;
      mxi.t[i] = itofix(7*i + 11);
   }
   build_matrix_al(&mxf, &mxal);

   xf = 3; xi = itofix(3);
   yf = 5; yi = itofix(5);
   zf = 7; zi = itofix(7);

   set_projection_viewport(0, 0, 640, 480);



   printf("\nApply_matrix\n");

   apply_matrix(&mxi, xi, yi, zi, &xio, &yio, &zio);
   printf("Fixed %12.6f %12.6f %12.6f\n", fixtof(xio), fixtof(yio), fixtof(zio));

   apply_matrix_f(&mxf, xf, yf, zf, &xfo, &yfo, &zfo);
   printf("Float %12.6f %12.6f %12.6f\n", xfo, yfo, zfo);

   if (cpu_3dnow) {
      apply_matrix_3dnow(&mxal, xf, yf, zf, &xfo, &yfo, &zfo);
      printf("3DNow %12.6f %12.6f %12.6f\n", xfo, yfo, zfo);
   }



   printf("\nPersp_project\n");

   persp_project(xi, yi, zi, &xio, &yio);
   printf("Fixed %12.6f %12.6f\n", fixtof(xio), fixtof(yio));

   persp_project_f(xf, yf, zf, &xfo, &yfo);
   printf("Float %12.6f %12.6f\n", xfo, yfo);

   if (cpu_3dnow) {
      persp_project_3dnow(xf, yf, zf, &xfo, &yfo);
      printf("Float %12.6f %12.6f\n", xfo, yfo);
   }



   printf("\nNormalize_vector\n");

   xio = xi; yio = yi; zio = zi;
   normalize_vector(&xio, &yio, &zio);
   printf("Fixed %12.8f %12.8f %12.8f\n", fixtof(xio), fixtof(yio), fixtof(zio));

   xfo = xf; yfo = yf; zfo = zf;
   normalize_vector_f(&xfo, &yfo, &zfo);
   printf("Float %12.8f %12.8f %12.8f\n", xfo, yfo, zfo);

   if (cpu_3dnow) {
      xfo = xf; yfo = yf; zfo = zf;
      normalize_vector_3dnow(&xfo, &yfo, &zfo);
      printf("3DNow %12.8f %12.8f %12.8f\n", xfo, yfo, zfo);
   }

   return 0;
}
