/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves
 *      shawn@talula.demon.co.uk
 *      http://www.talula.demon.co.uk/allegro/
 *
 *      Different MMX defines
 *
 *      This file written by Calin Andrian.
 *
 *      See readme.txt for copyright information.
 */


/* MMX defs */

#define MMX_REGS(src, dst)         .byte 0xc0 + 8 * dst + src
#define MMX_REGM(mem, reg)         .byte 5 + 8 * reg ; .long mem
#define MMX_REGSP(dis, reg)        .byte 0x44 + 8 * reg ; .byte 0x24, dis
#define MMX_REGBP(base, reg)       .byte 8 * reg + base
#define MMX_REGDP(base, dis, reg)  .byte 0x40 + 8 * reg + base ; .byte dis


/* PAND is broken in GAS 2.8.1 */

#define PAND_R(src, dst)  .byte 0x0f, 0xdb ; MMX_REGS(src, dst)
#define PAND_M(mem, reg)  .byte 0x0f, 0xdb ; MMX_REGM(mem, reg)
#define PAND_S(dis, reg)  .byte 0x0f, 0xdb ; MMX_REGSP(dis, reg)


/* 3Dnow! defs */
#define OP3D_R(src, dst, op) .byte 0x0f, 0x0f ; MMX_REGS(src, dst) ; .byte op
#define OP3D_S(dis, reg, op) .byte 0x0f, 0x0f ; MMX_REGSP(dis, reg) ; .byte op

#define PAVGUSB_R(src, dst)         OP3D_R(src, dst, 0xbf)
#define PF2ID_R(src, dst)           OP3D_R(src, dst, 0x1d)
#define PFACC_R(src, dst)           OP3D_R(src, dst, 0xae)
#define PFADD_R(src, dst)           OP3D_R(src, dst, 0x9e)
#define PFCMPEQ_R(src, dst)         OP3D_R(src, dst, 0xb0)
#define PFCMPGE_R(src, dst)         OP3D_R(src, dst, 0x90)
#define PFCMPGT_R(src, dst)         OP3D_R(src, dst, 0xa0)
#define PFMAX_R(src, dst)           OP3D_R(src, dst, 0xa4)
#define PFMIN_R(src, dst)           OP3D_R(src, dst, 0x94)
#define PFMUL_R(src, dst)           OP3D_R(src, dst, 0xb4)
#define PFRCP_R(src, dst)           OP3D_R(src, dst, 0x96)
#define PFRCPIT1_R(src, dst)        OP3D_R(src, dst, 0xa6)
#define PFRCPIT2_R(src, dst)        OP3D_R(src, dst, 0xb6)
#define PFRSQIT1_R(src, dst)        OP3D_R(src, dst, 0xa7)
#define PFRSQRT_R(src, dst)         OP3D_R(src, dst, 0x97)
#define PFSUB_R(src, dst)           OP3D_R(src, dst, 0x9a)
#define PFSUBR_R(src, dst)          OP3D_R(src, dst, 0xaa)
#define PI2FD_R(src, dst)           OP3D_R(src, dst, 0x0d)
#define PMULHRW_R(src, dst)         OP3D_R(src, dst, 0xb7)

#define PFADD_S(dis, reg)           OP3D_S(dis, reg, 0x9e)

#define FEMMS_R()                   .byte 0x0f, 0x0e
#define PREFETCH_B(base)            .byte 0x0f, 0x0d ; MMX_REGBP(base, 0x0)
#define PREFETCHW_B(base)           .byte 0x0f, 0x0d ; MMX_REGBP(base, 0x1)


/* KNI defs */

#define KNI_R(src, dst, op)       .byte 0x0f, op ; MMX_REGS(src, dst)
#define KNI_M(mem, reg, op)       .byte 0x0f, op ; MMX_REGM(mem, reg)
#define KNI_S(dis, reg, op)       .byte 0x0f, op ; MMX_REGSP(dis, reg)
#define KNI_B(base, reg, op)      .byte 0x0f, op ; MMX_REGBP(base, reg)
#define KNI_D(base, dis, reg, op) .byte 0x0f, op ; MMX_REGDP(base, dis, reg)

#define ADDPS_R(src, dst)                    KNI_R(src, dst, 0x58)
#define ADDSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x58)
#define ANDNPS_R(src, dst)                   KNI_R(src, dst, 0x55)
#define ANDPS_R(src, dst)                    KNI_R(src, dst, 0x54)
#define CMPPS_R(src, dst, imm)               KNI_R(src, dst, 0xc2) ; .byte imm
#define CMPSS_R(src, dst, imm)  .byte 0xf3 ; KNI_R(src, dst, 0xc2) ; .byte imm
#define COMISS_R(src, dst)                   KNI_R(src, dst, 0x2f)
#define CVTPI2PS_R(src, dst)                 KNI_R(src, dst, 0x2a)
#define CVTPS2PI_R(src, dst)                 KNI_R(src, dst, 0x2d)
#define CVTSI2SS_R(src, dst)    .byte 0xf3 ; KNI_R(src, dst, 0x2a)
#define CVTSS2SI_R(src, dst)    .byte 0xf3 ; KNI_R(src, dst, 0x2d)
#define CVTTPS2PI_R(src, dst)                KNI_R(src, dst, 0x2c)
#define CVTTSS2SI_R(src, dst)   .byte 0xf3 ; KNI_R(src, dst, 0x2c)
#define DIVPS_R(src, dst)                    KNI_R(src, dst, 0x5e)
#define DIVSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x5e)
#define MAXPS_R(src, dst)                    KNI_R(src, dst, 0x5f)
#define MAXSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x5f)
#define MINPS_R(src, dst)                    KNI_R(src, dst, 0x5d)
#define MINSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x5d)
#define MOVAPS_R(src, dst)                   KNI_R(src, dst, 0x28)
#define MOVHLPS_R(src, dst)                  KNI_R(src, dst, 0x12)
#define MOVLHPS_R(src, dst)                  KNI_R(src, dst, 0x16)
#define MOVMSKPS_R(src, dst)                 KNI_R(src, dst, 0x50)
#define MOVSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x10)
#define MOVUPS_R(src, dst)                   KNI_R(src, dst, 0x10)
#define MULPS_R(src, dst)                    KNI_R(src, dst, 0x59)
#define MULSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x59)
#define ORPS_R(src, dst)                     KNI_R(src, dst, 0x56)
#define RCPPS_R(src, dst)                    KNI_R(src, dst, 0x53)
#define RCPSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x53)
#define RSQRTPS_R(src, dst)                  KNI_R(src, dst, 0x52)
#define RSQRTSS_R(src, dst)     .byte 0xf3 ; KNI_R(src, dst, 0x52)
#define SHUFPS_R(src, dst, imm)              KNI_R(src, dst, 0xc6) ; .byte imm
#define SQRTPS_R(src, dst)                   KNI_R(src, dst, 0x51)
#define SQRTSS_R(src, dst)      .byte 0xf3 ; KNI_R(src, dst, 0x51)
#define SUBPS_R(src, dst)                    KNI_R(src, dst, 0x5c)
#define SUBSS_R(src, dst)       .byte 0xf3 ; KNI_R(src, dst, 0x5c)
#define UCOMISS_R(src, dst)                  KNI_R(src, dst, 0x2e)
#define UNPCKHPS_R(src, dst)                 KNI_R(src, dst, 0x15)
#define UNPCKLPS_R(src, dst)                 KNI_R(src, dst, 0x14)
#define XORPS_R(src, dst)                    KNI_R(src, dst, 0x57)

#define PAVGB_R(src, dst)                    KNI_R(src, dst, 0xe0)
#define PAVGW_R(src, dst)                    KNI_R(src, dst, 0xe3)
#define PEXTRW_R(src, dst, imm)              KNI_R(src, dst, 0xc5) ; .byte imm
#define PINSRW_R(src, dst, imm)              KNI_R(src, dst, 0xc4) ; .byte imm
#define PMAXSW_R(src, dst)                   KNI_R(src, dst, 0xee)
#define PMAXUB_R(src, dst)                   KNI_R(src, dst, 0xde)
#define PMINSW_R(src, dst)                   KNI_R(src, dst, 0xea)
#define PMINUB_R(src, dst)                   KNI_R(src, dst, 0xda)
#define PMOVMSKB_R(src, dst)                 KNI_R(src, dst, 0xd7)
#define PMULHUW_R(src, dst)                  KNI_R(src, dst, 0xe4)
#define PSADBW_R(src, dst)                   KNI_R(src, dst, 0xf6)
#define PSHUFW_R(src, dst, imm)              KNI_R(src, dst, 0x70) ; .byte imm
#define MASKMOVQ_R(src, dst)                 KNI_R(src, dst, 0xf7)

#define MOVAPS_LD_S(dis, reg)                KNI_S(dis, reg, 0x28)
#define MOVAPS_ST_S(reg, dis)                KNI_S(dis, reg, 0x29)
#define MOVHPS_LD_S(dis, reg)                KNI_S(dis, reg, 0x16)
#define MOVHPS_ST_S(reg, dis)                KNI_S(dis, reg, 0x17)
#define MOVLPS_LD_S(dis, reg)                KNI_S(dis, reg, 0x12)
#define MOVLPS_ST_S(reg, dis)                KNI_S(dis, reg, 0x13)
#define MOVSS_LD_S(dis, reg)    .byte 0xf3 ; KNI_S(dis, reg, 0x10)
#define MOVSS_ST_S(reg, dis)    .byte 0xf3 ; KNI_S(dis, reg, 0x11)
#define MOVNTPS_S(reg, dis)                  KNI_S(dis, reg, 0x2b)
#define MOVNTQ_S(reg, dis)                   KNI_S(dis, reg, 0xe7)

#define FXRSTOR_S(dis)                       KNI_S(dis, 0x1, 0xae)
#define FXSAVE_S(dis)                        KNI_S(dis, 0x0, 0xae)
#define LDMXCSR_S(dis)                       KNI_S(dis, 0x2, 0xae)
#define STMXCSR_S(dis)                       KNI_S(dis, 0x3, 0xae)
#define PREFETCH0_B(base)                    KNI_B(base, 0x1, 0x18)
#define PREFETCH1_B(base)                    KNI_B(base, 0x2, 0x18)
#define PREFETCH2_B(base)                    KNI_B(base, 0x3, 0x18)
#define PREFETCHNTA_B(base)                  KNI_B(base, 0x0, 0x18)
#define SFENCE()                .byte 0x0f, 0xae, 0xf8

#define MOVAPS_LD_M(mem, reg)                   KNI_M(mem, reg, 0x28)
#define MOVAPS_ST_M(reg, mem)                   KNI_M(mem, reg, 0x29)
#define MOVUPS_LD_M(mem, reg)                   KNI_M(mem, reg, 0x10)
#define MOVUPS_ST_M(reg, mem)                   KNI_M(mem, reg, 0x11)
#define MOVSS_LD_M(mem, reg)       .byte 0xf3 ; KNI_M(mem, reg, 0x10)
#define MOVSS_ST_M(reg, mem)       .byte 0xf3 ; KNI_M(mem, reg, 0x11)

#define MOVAPS_LD_B(base, reg)                  KNI_B(base, reg, 0x28)
#define MOVAPS_ST_B(reg, base)                  KNI_B(base, reg, 0x29)
#define MOVUPS_LD_B(base, reg)                  KNI_B(base, reg, 0x10)
#define MOVUPS_ST_B(reg, base)                  KNI_B(base, reg, 0x11)
#define MOVSS_LD_B(base, reg)      .byte 0xf3 ; KNI_B(base, reg, 0x10)
#define MOVSS_ST_B(reg, base)      .byte 0xf3 ; KNI_B(base, reg, 0x11)

#define MOVAPS_LD_D(base, dis, reg)             KNI_D(base, dis, reg, 0x28)
#define MOVAPS_ST_D(reg, base, dis)             KNI_D(base, dis, reg, 0x29)
#define MOVUPS_LD_D(base, dis, reg)             KNI_D(base, dis, reg, 0x10)
#define MOVUPS_ST_D(reg, base, dis)             KNI_D(base, dis, reg, 0x11)
#define MOVSS_LD_D(base, dis, reg) .byte 0xf3 ; KNI_D(base, dis, reg, 0x10)
#define MOVSS_ST_D(reg, base, dis) .byte 0xf3 ; KNI_D(base, dis, reg, 0x11)

#define CMPEQ           0
#define CMPLT           1
#define CMPLE           2
#define CMPUNORD        3
#define CMPNEQ          4
#define CMPNLT          5
#define CMPNLE          6
#define CMPORD          7

