/*         ______   ___    ___ 
 *        /\  _  \ /\_ \  /\_ \ 
 *        \ \ \L\ \\//\ \ \//\ \      __     __   _ __   ___ 
 *         \ \  __ \ \ \ \  \ \ \   /'__`\ /'_ `\/\`'__\/ __`\
 *          \ \ \/\ \ \_\ \_ \_\ \_/\  __//\ \L\ \ \ \//\ \L\ \
 *           \ \_\ \_\/\____\/\____\ \____\ \____ \ \_\\ \____/
 *            \/_/\/_/\/____/\/____/\/____/\/___L\ \/_/ \/___/
 *                                           /\____/
 *                                           \_/__/
 *      By Shawn Hargreaves
 *      shawn@talula.demon.co.uk
 *      http://www.talula.demon.co.uk/allegro/
 *
 *      MMX versions of the polygon scanline fillers
 *
 *      This file written by Calin Andrian.
 *
 *      See readme.txt for copyright information.
 */

#include "asmdefs.inc"
#include "mmxdefs.inc"
#include "mmx.h"

#ifdef ALLEGRO_MMX

.data

		    .align 4
__mask_mmx_15:      .long 0x03e0001f, 0x007c
__mask_mmx_16:      .long 0x07e0001f, 0x00f8
__zero_mmx:         .long 0, 0


.text



/* all these functions share the same parameters */

#define ADDR      ARG1
#define W         ARG2
#define INFO      ARG3


#define TMP    12(%esp)
#define TMP4   16(%esp)
#define DB     20(%esp)

.globl __poly_scanline_grgb8x
   .align 4
__poly_scanline_grgb8x:
   pushl %ebp
   movl %esp, %ebp
   subl $20, %esp
   andl $-8, %esp

   pushl %ebx
   pushl %esi
   pushl %edi

   movl INFO, %esi               /* load registers */

   movl POLYSEG_R(%esi), %eax
   movl %eax, TMP
   movl POLYSEG_DR(%esi), %edi
   addl %edi, %eax
   movl %eax, TMP4
   shll $1, %edi
   movq TMP, %mm0
   movd %edi, %mm1
   punpckldq %mm1, %mm1

   movl POLYSEG_G(%esi), %eax
   movl %eax, TMP
   movl POLYSEG_DG(%esi), %edi
   addl %edi, %eax
   movl %eax, TMP4
   shll $1, %edi
   movq TMP, %mm2
   movd %edi, %mm3
   punpckldq %mm3, %mm3

   movl POLYSEG_B(%esi), %eax
   movl %eax, TMP
   movl POLYSEG_DB(%esi), %edi
   addl %edi, %eax
   movl %eax, TMP4
   shll $1, %edi
   movq TMP, %mm4
   movd %edi, %mm5
   punpckldq %mm5, %mm5
   movq %mm5, DB

   movl _rgb_map, %esi
   movl ADDR, %edi
   movl W, %ecx

   .align 4
grgb_loop_x:
   movq %mm0, %mm7              /* red */
   movq %mm2, %mm6              /* green */
   psrld $19, %mm7
   movq %mm4, %mm5              /* blue */
   psrld $19, %mm6
   psrld $19, %mm5
   pslld $10, %mm7
   pslld $5, %mm6
   por %mm5, %mm7
   por %mm6, %mm7

   subl $2, %ecx
   movd %mm7, %edx
   movb (%esi, %edx), %al        /* table lookup */
   jl grgb_last_x

   psrlq $32, %mm7
   movd %mm7, %edx
   movb (%esi, %edx), %ah        /* table lookup */

   paddd %mm1, %mm0
   paddd %mm3, %mm2
   paddd DB, %mm4

   movw %ax, %fs:(%edi)          /* write two pixels */
   leal 2(%edi), %edi
   jg grgb_loop_x

grgb_done_x:
   emms
   popl %edi
   popl %esi
   popl %ebx
   movl %ebp, %esp
   popl %ebp
   ret                           /* end of _poly_scanline_grgb8x() */
   .align 4
grgb_last_x:
   movb %al, %fs:(%edi)          /* write last pixel */
   jmp grgb_done_x

#undef TMP
#undef TMP4
#undef DB



#define TMP    12(%esp)
#define TMP4   16(%esp)
#define SR     20(%esp)
#define SG     28(%esp)
#define SB     36(%esp)
#define DB     44(%esp)

/* first part of MMX grgb routine - two for the price of one */
#define INIT_GRGB_MMX(depth, r_sft, g_sft, b_sft)                     \
   pushl %ebp                                                       ; \
   movl %esp, %ebp                                                  ; \
   subl $44, %esp                                                   ; \
   andl $-8, %esp                                                   ; \
   pushl %ebx                                                       ; \
   pushl %esi                                                       ; \
   pushl %edi                                                       ; \
								    ; \
   movl INFO, %esi               /* load registers */               ; \
								    ; \
   movl POLYSEG_R(%esi), %eax                                       ; \
   movl %eax, TMP                                                   ; \
   movl POLYSEG_DR(%esi), %edi                                      ; \
   addl %edi, %eax                                                  ; \
   movl %eax, TMP4                                                  ; \
   shll $1, %edi                                                    ; \
   movq TMP, %mm0                                                   ; \
   movd %edi, %mm1                                                  ; \
   movd __rgb_r_shift_##depth, %mm7                                 ; \
   punpckldq %mm1, %mm1                                             ; \
   movq %mm7, SR                                                    ; \
								    ; \
   movl POLYSEG_G(%esi), %eax                                       ; \
   movl %eax, TMP                                                   ; \
   movl POLYSEG_DG(%esi), %edi                                      ; \
   addl %edi, %eax                                                  ; \
   movl %eax, TMP4                                                  ; \
   shll $1, %edi                                                    ; \
   movq TMP, %mm2                                                   ; \
   movd %edi, %mm3                                                  ; \
   movd __rgb_g_shift_##depth, %mm7                                 ; \
   punpckldq %mm3, %mm3                                             ; \
   movq %mm7, SG                                                    ; \
								    ; \
   movl POLYSEG_B(%esi), %eax                                       ; \
   movl %eax, TMP                                                   ; \
   movl POLYSEG_DB(%esi), %edi                                      ; \
   addl %edi, %eax                                                  ; \
   movl %eax, TMP4                                                  ; \
   shll $1, %edi                                                    ; \
   movq TMP, %mm4                                                   ; \
   movd %edi, %mm5                                                  ; \
   movd __rgb_b_shift_##depth, %mm7                                 ; \
   punpckldq %mm5, %mm5                                             ; \
   movq %mm7, SB                                                    ; \
   movq %mm5, DB                                                    ; \
								    ; \
   movl ADDR, %edi                                                  ; \
   movl W, %ecx                                                     ; \
								    ; \
   .align 4                                                         ; \
grgb_loop_x_##depth:                                                ; \
   movq %mm0, %mm7              /* red */                           ; \
   movq %mm2, %mm6              /* green */                         ; \
   psrld $r_sft, %mm7                                               ; \
   movq %mm4, %mm5              /* blue */                          ; \
   psrld $g_sft, %mm6                                               ; \
   psrld $b_sft, %mm5                                               ; \
   pslld SR, %mm7                                                   ; \
   pslld SG, %mm6                                                   ; \
   pslld SB, %mm5                                                   ; \
   por %mm6, %mm7                                                   ; \
   por %mm5, %mm7

/* end of grgb MMX routine */
#define END_GRGB_MMX(depth)                                           \
   paddd %mm1, %mm0                                                 ; \
   paddd %mm3, %mm2                                                 ; \
   paddd DB, %mm4                                                   ; \
   jg grgb_loop_x_##depth       /* watch out for flags */           ; \
99:                                                                 ; \
   emms                                                             ; \
   popl %edi                                                        ; \
   popl %esi                                                        ; \
   popl %ebx                                                        ; \
   movl %ebp, %esp                                                  ; \
   popl %ebp



#ifdef ALLEGRO_COLOR16
.globl __poly_scanline_grgb15x
   .align 4
__poly_scanline_grgb15x:
   INIT_GRGB_MMX(15, 19, 19, 19)
   subl $2, %ecx
   jl 7f                        /* from here on, flags don't change */
   packssdw %mm7, %mm7
   movd %mm7, %fs:(%edi)        /* write two pixels */
   leal 4(%edi), %edi
   END_GRGB_MMX(15)
   ret                           /* end of _poly_scanline_grgb15x() */
   .align 4
7:
   movd %mm7, %eax
   movw %ax, %fs:(%edi)         /* write the pixel */
   jmp 99b

.globl __poly_scanline_grgb16x
   .align 4
__poly_scanline_grgb16x:
   INIT_GRGB_MMX(16, 19, 18, 19)
   subl $2, %ecx
   jl 7f
   movd %mm7, %edx
   psrlq $16, %mm7
   movd %mm7, %eax
   movw %dx, %ax
   movl %eax, %fs:(%edi)        /* write two pixels */
   leal 4(%edi), %edi
   END_GRGB_MMX(16)
   ret                           /* end of _poly_scanline_grgb16x() */
   .align 4
7:
   movd %mm7, %eax
   movw %ax, %fs:(%edi)         /* write the pixel */
   jmp 99b
#endif /* COLOR16 */



#ifdef ALLEGRO_COLOR32
.globl __poly_scanline_grgb32x
   .align 4
__poly_scanline_grgb32x:
   INIT_GRGB_MMX(32, 16, 16, 16)
   subl $2, %ecx
   jl 7f
   movq %mm7, %fs:(%edi)         /* write two pixels */
   leal 8(%edi), %edi
   END_GRGB_MMX(32)
   ret                           /* end of _poly_scanline_grgb32x() */
   .align 4
7:
   movd %mm7, %fs:(%edi)         /* write the pixel */
   jmp 99b
#endif /* COLOR32 */

#ifdef ALLEGRO_COLOR24
.globl __poly_scanline_grgb24x
   .align 4
__poly_scanline_grgb24x:
   INIT_GRGB_MMX(24, 16, 16, 16)
   movd %mm7, %eax
   movw %ax, %fs:(%edi)          /* write the pixel */
   shrl $16, %eax
   movb %al, %fs:2(%edi)
   subl $2, %ecx
   jl 99f
   psrlq $32, %mm7
   movd %mm7, %eax
   movw %ax, %fs:3(%edi)         /* write second pixel */
   shrl $16, %eax
   movb %al, %fs:5(%edi)
   leal 6(%edi), %edi
   cmpl $0, %ecx
   END_GRGB_MMX(24)
   ret                           /* end of _poly_scanline_grgb24x() */
#endif /* COLOR24 */

#undef TMP
#undef TMP2
#undef TMP4
#undef TMP6
#undef SR
#undef SG
#undef SB



#define DV       12(%esp)
#define DV4      16(%esp)
#define DU       20(%esp)
#define DU4      24(%esp)
#define VSHIFT   28(%esp)
#define VSHIFT4  32(%esp)
#define DALPHA   36(%esp)
#define BLEND    44(%esp)
#define TMP      52(%esp)
#define TMP2     54(%esp)
#define TMP4     56(%esp)

/* first part of an affine texture mapping operation */
#define INIT_ATEX_MMX(extra...)                                       \
   pushl %ebp                                                       ; \
   movl %esp, %ebp                                                  ; \
   subl $56, %esp                    /* local variables */          ; \
   andl $-8, %esp                                                   ; \
   pushl %ebx                                                       ; \
   pushl %esi                                                       ; \
   pushl %edi                                                       ; \
								    ; \
   movl INFO, %esi               /* load registers */               ; \
   extra                                                            ; \
								    ; \
   movl POLYSEG_VSHIFT(%esi), %ecx                                  ; \
   movl POLYSEG_VMASK(%esi), %eax                                   ; \
   shll %cl, %eax           /* adjust v mask and shift value */     ; \
   negl %ecx                                                        ; \
   addl $16, %ecx                                                   ; \
   movd %eax, %mm2                                                  ; \
   movl %ecx, VSHIFT                                                ; \
   punpckldq %mm2, %mm2                                             ; \
   movl $0, VSHIFT4                                                 ; \
								    ; \
   movd POLYSEG_UMASK(%esi), %mm3                                   ; \
   punpckldq %mm3, %mm3                                             ; \
                                                                    ; \
   movl POLYSEG_V(%esi), %eax                                       ; \
   movl POLYSEG_DV(%esi), %edx                                      ; \
   movl %eax, TMP                                                   ; \
   addl %edx, %eax                                                  ; \
   addl %edx, %edx                                                  ; \
   movl %eax, TMP4                                                  ; \
   movl %edx, DV                                                    ; \
   movq TMP, %mm0                                                   ; \
   movl %edx, DV4                                                   ; \
                                                                    ; \
   movl POLYSEG_U(%esi), %eax                                       ; \
   movl POLYSEG_DU(%esi), %edx                                      ; \
   movl %eax, TMP                                                   ; \
   addl %edx, %eax                                                  ; \
   addl %edx, %edx                                                  ; \
   movl %eax, TMP4                                                  ; \
   movl %edx, DU                                                    ; \
   movq TMP, %mm1                                                   ; \
   movl %edx, DU4                                                   ; \
                                                                    ; \
   movl ADDR, %edi                                                  ; \
   movl POLYSEG_TEXTURE(%esi), %esi                                 ; \
   movl W, %ecx                                                     ; \
								    ; \
   .align 4                                                         ; \
1:                                                                  ; \
   movq %mm0, %mm4                                                  ; \
   movq %mm1, %mm5                                                  ; \
   psrad VSHIFT, %mm4                                               ; \
   paddd DV, %mm0                                                   ; \
   psrad $16, %mm5                                                  ; \
   paddd DU, %mm1                                                   ; \
   PAND_R(2, 4)                                                     ; \
   PAND_R(3, 5)                                                     ; \
   paddd %mm5, %mm4                                                 ; \
   subl $2, %ecx                                                    ; \
   movd %mm4, %ebx                                                  ; \
   jl 77f                                                           ; \
   psrlq $32, %mm4                                                  ; \
   movd %mm4, %edx

/* end of atex routine */
#define END_ATEX_MMX()                                              ; \
   cmpl $0, %ecx                                                    ; \
   jg 1b                                                            ; \
99:                                                                 ; \
   emms                                                             ; \
   popl %edi                                                        ; \
   popl %esi                                                        ; \
   popl %ebx                                                        ; \
   movl %ebp, %esp                                                  ; \
   popl %ebp



/* helper for setting up ALPHA and DALPHA before MMX lit routines */
#define INIT_MMX_ALPHA()                                              \
   movd POLYSEG_DC(%esi), %mm6                                      ; \
   movd POLYSEG_C(%esi), %mm7                                       ; \
   punpckldq %mm6, %mm6                                             ; \
   punpckldq %mm7, %mm7                                             ; \
   movq %mm6, DALPHA



#define INIT_HIGH(depth)                                              \
   INIT_MMX_ALPHA()                                                 ; \
   movl __blender_col_##depth, %eax                                 ; \
   movw %ax, TMP                                                    ; \
   movw %ax, TMP2                                                   ; \
   movb %ah, TMP4                                                   ; \
   movq TMP, %mm4                                                   ; \
   PAND_M(__mask_mmx_##depth, 4)            /* pand mem, %mm4 */    ; \
   movq %mm4, BLEND

#define TWO_HIGH(depth)                                               \
   movw (%esi, %ebx, 2), %ax                                        ; \
   movw (%esi, %edx, 2), %bx                                        ; \
   movd %eax, %mm4                                                  ; \
   movd %ebx, %mm5                                                  ; \
                                                                    ; \
   punpcklwd %mm4, %mm4         /* 0 0 r1g1b1 r1g1b1 */             ; \
   punpcklwd %mm5, %mm5         /* 0 0 r2g2b2 r2g2b2 */             ; \
   punpckldq %mm5, %mm4         /* r2g2b2 r2g2b2 r1g1b1 r1g1b1 */   ; \
   movq %mm4, %mm6                                                  ; \
   movq %mm4, %mm5                                                  ; \
   psrlw $8, %mm6               /* r2 r2 r1 r1 */                   ; \
   punpckldq %mm6, %mm4         /* r1 r1 g1 b1 */                   ; \
   punpckhdq %mm6, %mm5         /* r2 r2 g2 b2 */                   ; \
                                                                    ; \
   movq %mm7, %mm6                                                  ; \
   PAND_M(__mask_mmx_##depth, 4)            /* pand mem, %mm4 */    ; \
   psrld $9, %mm6               /* 15-bit alpha */                  ; \
   psubw BLEND, %mm4                                                ; \
   packssdw %mm6, %mm6          /* 4 alphas */                      ; \
   paddw %mm4, %mm4                                                 ; \
   paddd DALPHA, %mm7                                               ; \
   pmulhw %mm6, %mm4                                                ; \
                                                                    ; \
   movq %mm7, %mm6                                                  ; \
   PAND_M(__mask_mmx_##depth, 5)            /* pand mem, %mm5 */    ; \
   psrld $9, %mm6                                                   ; \
   psubw BLEND, %mm5                                                ; \
   packssdw %mm6, %mm6                                              ; \
   paddw %mm5, %mm5                                                 ; \
   paddd DALPHA, %mm7                                               ; \
   pmulhw %mm6, %mm5                                                ; \
                                                                    ; \
   paddw BLEND, %mm4                                                ; \
   paddw BLEND, %mm5                                                ; \
   PAND_M(__mask_mmx_##depth, 4)            /* pand mem, %mm4 */    ; \
   PAND_M(__mask_mmx_##depth, 5)            /* pand mem, %mm5 */    ; \
                                                                    ; \
   movq %mm4, %mm6                                                  ; \
   punpckhwd %mm5, %mm4         /*  0  0 r2 r1 */                   ; \
   punpcklwd %mm5, %mm6         /* g2 g1 b2 b1 */                   ; \
   psllw $8, %mm4               /* shift back */                    ; \
   por %mm6, %mm4               /* g2 g1 r2b2 r1b1 */               ; \
   psrlq $32, %mm6              /* 0 0 g2 g1 */                     ; \
   por %mm6, %mm4               /* g2 g1 r2g2b2 r1g1b1 */           ; \
   movd %mm4, %fs:(%edi)                                            ; \
   addl $4, %edi

#define ONE_HIGH(depth)                                               \
   movw %ax, TMP                                                    ; \
   movw %ax, TMP2                                                   ; \
   movb %ah, TMP4                                                   ; \
   movq TMP, %mm4                                                   ; \
                                                                    ; \
   movq %mm7, %mm6                                                  ; \
   PAND_M(__mask_mmx_##depth, 4)            /* pand mem, %mm4 */    ; \
   psrld $9, %mm6               /* 15-bit alpha */                  ; \
   psubw BLEND, %mm4                                                ; \
   packssdw %mm6, %mm6          /* 4 alphas */                      ; \
   paddw %mm4, %mm4                                                 ; \
   pmulhw %mm6, %mm4                                                ; \
                                                                    ; \
   paddw BLEND, %mm4                                                ; \
   PAND_M(__mask_mmx_##depth, 4)            /* pand mem, %mm4 */    ; \
                                                                    ; \
   movq %mm4, TMP                                                   ; \
   movd %mm4, %eax                                                  ; \
   orw TMP2, %ax                                                    ; \
   orb TMP4, %ah                                                    ; \
   movw %ax, %fs:(%edi)

#define MASK_HIGH(depth)                                              \
   movw (%esi, %ebx, 2), %ax                                        ; \
   cmpw $MASK_COLOR_##depth, %ax                                    ; \
   jz 7f                                                            ; \
   ONE_HIGH(depth)                                                  ; \
7:                                                                  ; \
   paddd DALPHA, %mm7                                               ; \
   addl $2, %edi



#define INIT_TRUE(depth)                                              \
   INIT_MMX_ALPHA()                                                 ; \
   movd __blender_col_##depth, %mm4                                 ; \
   punpcklbw __zero_mmx, %mm4                                       ; \
   movq %mm4, BLEND

#define TWO_TRUE()                                                    \
   movq %mm7, %mm6                                                  ; \
   punpcklbw __zero_mmx, %mm4                                       ; \
   psrld $9, %mm6               /* 15-bit alpha */                  ; \
   psubw BLEND, %mm4                                                ; \
   packssdw %mm6, %mm6          /* 4 alphas */                      ; \
   paddw %mm4, %mm4                                                 ; \
   paddd DALPHA, %mm7                                               ; \
   pmulhw %mm6, %mm4                                                ; \
                                                                    ; \
   movq %mm7, %mm6                                                  ; \
   punpcklbw __zero_mmx, %mm5                                       ; \
   psrld $9, %mm6                                                   ; \
   psubw BLEND, %mm5                                                ; \
   packssdw %mm6, %mm6                                              ; \
   paddw %mm5, %mm5                                                 ; \
   paddd DALPHA, %mm7                                               ; \
   pmulhw %mm6, %mm5                                                ; \
                                                                    ; \
   paddw BLEND, %mm4                                                ; \
   paddw BLEND, %mm5

#define ONE_TRUE()                                                    \
   movq %mm7, %mm6                                                  ; \
   punpcklbw __zero_mmx, %mm4                                       ; \
   psrld $9, %mm6               /* 15-bit alpha */                  ; \
   psubw BLEND, %mm4                                                ; \
   packssdw %mm6, %mm6          /* 4 alphas */                      ; \
   paddw %mm4, %mm4                                                 ; \
   pmulhw %mm6, %mm4                                                ; \
   paddw BLEND, %mm4

#define TWO_32()                                                      \
   movd (%esi, %ebx, 4), %mm4                                       ; \
   movd (%esi, %edx, 4), %mm5                                       ; \
   TWO_TRUE()                                                       ; \
   packuswb %mm5, %mm4                                              ; \
   movq %mm4, %fs:(%edi)                                            ; \
   addl $8, %edi

#define ONE_32()                                                      \
   movd (%esi, %ebx, 4), %mm4                                       ; \
   ONE_TRUE()                                                       ; \
   packuswb %mm4, %mm4                                              ; \
   movd %mm4, %fs:(%edi)

#define MASK_32()                                                     \
   movl (%esi, %ebx, 4), %eax                                       ; \
   cmpl $MASK_COLOR_32, %eax                                        ; \
   jz 7f                                                            ; \
   movd %eax, %mm4                                                  ; \
   ONE_TRUE()                                                       ; \
   packuswb %mm4, %mm4                                              ; \
   movd %mm4, %fs:(%edi)                                            ; \
7:                                                                  ; \
   paddd DALPHA, %mm7                                               ; \
   addl $4, %edi

#define TWO_24()                                                      \
   leal (%ebx, %ebx, 2), %ebx                                       ; \
   leal (%edx, %edx, 2), %edx                                       ; \
   movb 2(%esi, %ebx), %al                                          ; \
   shll $16, %eax                                                   ; \
   movw (%esi, %ebx), %ax                                           ; \
   movb 2(%esi, %edx), %bl                                          ; \
   shll $16, %ebx                                                   ; \
   movw (%esi, %edx), %bx                                           ; \
   movd %eax, %mm4                                                  ; \
   movd %ebx, %mm5                                                  ; \
   TWO_TRUE()                                                       ; \
   psllq $16, %mm4                                                  ; \
   packuswb %mm5, %mm4                                              ; \
   psrlq $8, %mm4                                                   ; \
   movd %mm4, %fs:(%edi)                                            ; \
   psrlq $32, %mm4                                                  ; \
   movd %mm4, %eax                                                  ; \
   movw %ax, %fs:4(%edi)                                            ; \
   addl $6, %edi

#define ONE_24()                                                      \
   leal (%ebx, %ebx, 2), %ebx                                       ; \
   movb 2(%esi, %ebx), %al                                          ; \
   shll $16, %eax                                                   ; \
   movw (%esi, %ebx), %ax                                           ; \
   movd %eax, %mm4                                                  ; \
   ONE_TRUE()                                                       ; \
   packuswb %mm4, %mm4                                              ; \
   movd %mm4, %eax                                                  ; \
   movw %ax, %fs:(%edi)                                             ; \
   shrl $16, %eax                                                   ; \
   movb %al, %fs:2(%edi)

#define MASK_24()                                                     \
   leal (%ebx, %ebx, 2), %ebx                                       ; \
   movb 2(%esi, %ebx), %al                                          ; \
   shll $16, %eax                                                   ; \
   movw (%esi, %ebx), %ax                                           ; \
   cmpl $MASK_COLOR_24, %eax                                        ; \
   jz 7f                                                            ; \
   movd %eax, %mm4                                                  ; \
   ONE_TRUE()                                                       ; \
   packuswb %mm4, %mm4                                              ; \
   movd %mm4, %eax                                                  ; \
   movw %ax, %fs:(%edi)                                             ; \
   shrl $16, %eax                                                   ; \
   movb %al, %fs:2(%edi)                                            ; \
7:                                                                  ; \
   paddd DALPHA, %mm7                                               ; \
   addl $3, %edi



#ifdef ALLEGRO_COLOR16
.globl __poly_scanline_atex_lit15x
   .align 4
__poly_scanline_atex_lit15x:
   INIT_ATEX_MMX(
     INIT_HIGH(15)
   )
   TWO_HIGH(15)
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_lit15x() */
   .align 4
77:
   movw (%esi, %ebx, 2), %ax
   ONE_HIGH(15)
   jmp 99b

.globl __poly_scanline_atex_mask_lit15x
   .align 4
__poly_scanline_atex_mask_lit15x:
   INIT_ATEX_MMX(
     INIT_HIGH(15)
   )
   MASK_HIGH(15)
   movl %edx, %ebx
77:
   MASK_HIGH(15)
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_mask_lit15x() */



.globl __poly_scanline_atex_lit16x
   .align 4
__poly_scanline_atex_lit16x:
   INIT_ATEX_MMX(
     INIT_HIGH(16)
   )
   TWO_HIGH(16)
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_lit16x() */
   .align 4
77:
   movw (%esi, %ebx, 2), %ax
   ONE_HIGH(16)
   jmp 99b

.globl __poly_scanline_atex_mask_lit16x
   .align 4
__poly_scanline_atex_mask_lit16x:
   INIT_ATEX_MMX(
     INIT_HIGH(16)
   )
   MASK_HIGH(16)
   movl %edx, %ebx
77:
   MASK_HIGH(16)
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_mask_lit16x() */
#endif /* COLOR16 */



#ifdef ALLEGRO_COLOR32
.globl __poly_scanline_atex_lit32x
   .align 4
__poly_scanline_atex_lit32x:
   INIT_ATEX_MMX(
     INIT_TRUE(32)
   )
   TWO_32()
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_lit32x() */
   .align 4
77:
   ONE_32()
   jmp 99b

.globl __poly_scanline_atex_mask_lit32x
   .align 4
__poly_scanline_atex_mask_lit32x:
   INIT_ATEX_MMX(
     INIT_TRUE(32)
   )
   MASK_32()
   movl %edx, %ebx
77:
   MASK_32()
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_mask_lit32x() */
#endif /* COLOR32 */



#ifdef ALLEGRO_COLOR24
.globl __poly_scanline_atex_lit24x
   .align 4
__poly_scanline_atex_lit24x:
   INIT_ATEX_MMX(
     INIT_TRUE(24)
   )
   TWO_24()
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_lit24x() */
   .align 4
77:
   ONE_24()
   jmp 99b

.globl __poly_scanline_atex_mask_lit24x
   .align 4
__poly_scanline_atex_mask_lit24x:
   INIT_ATEX_MMX(
     INIT_TRUE(24)
   )
   MASK_24()
   movl %edx, %ebx
77:
   MASK_24()
   END_ATEX_MMX()
   ret                           /* end of _poly_scanline_atex_mask_lit24x() */
#endif /* COLOR24 */



#define VMASK    60(%esp)
#define VMASK4   64(%esp)
#define UMASK    68(%esp)
#define UMASK4   72(%esp)
#define ALPHA    76(%esp)
#define ALPHA2   78(%esp)
#define ALPHA4   80(%esp)
#define V1       84(%esp)
#define U1       88(%esp)
#define DFV      92(%esp)
#define DFU      96(%esp)
#define DFZ     100(%esp)
#define V       108(%esp)
#define U       112(%esp)
#define SM      116(%esp)
#define SV      120(%esp)
#define SU      124(%esp)
#define SZ      128(%esp)

#define VMASK_N  60
#define UMASK_N  68
#define DFV_N    92
#define DFZ_N   100

/* main body of the perspective-correct texture mapping routine,
   using 3D enhancement CPUs */
#define INIT_PTEX_3D(extra...)                                        \
   pushl %ebp                                                       ; \
   movl %esp, %ebp                                                  ; \
   subl $128, %esp               /* local variables */              ; \
   andl $-8, %esp                                                   ; \
   pushl %ebx                                                       ; \
   pushl %esi                                                       ; \
   pushl %edi                                                       ; \
								    ; \
   FEMMS_R()                                                        ; \
   movl INFO, %esi               /* load registers */               ; \
								    ; \
   extra                                                            ; \
								    ; \
   movd POLYSEG_DZ(%esi), %mm1                                      ; \
   movd POLYSEG_DFV(%esi), %mm2                                     ; \
   movd POLYSEG_DFU(%esi), %mm3                                     ; \
   punpckldq %mm3, %mm2                                             ; \
   PFADD_R(1, 1)                /* pfadd %mm1, %mm1 */              ; \
   movq %mm1, DFZ                                                   ; \
   PFADD_R(2, 2)                /* pfadd %mm2, %mm2 */              ; \
   movq %mm2, DFV                                                   ; \
								    ; \
   movd POLYSEG_Z(%esi), %mm0                                       ; \
   movd POLYSEG_FV(%esi), %mm1                                      ; \
   movd POLYSEG_FU(%esi), %mm3                                      ; \
   PFRCP_R(0, 2)                /* pfrcp %mm0, %mm2 */              ; \
   punpckldq %mm3, %mm1                                             ; \
   PFMUL_R(1, 2)                /* pfmul %mm1, %mm2 */              ; \
   PF2ID_R(2, 2)                /* pf2id %mm2, %mm2 */              ; \
								    ; \
   movl POLYSEG_VSHIFT(%esi), %ecx                                  ; \
   movl POLYSEG_VMASK(%esi), %eax                                   ; \
   movl POLYSEG_UMASK(%esi), %edx                                   ; \
   shll %cl, %eax           /* adjust v mask and shift value */     ; \
   negl %ecx                                                        ; \
   addl $16, %ecx                                                   ; \
   movl %ecx, VSHIFT                                                ; \
   movl $0, VSHIFT4                                                 ; \
   movl %eax, VMASK                                                 ; \
   movl %eax, VMASK4                                                ; \
   movl %edx, UMASK                                                 ; \
   movl %edx, UMASK4                                                ; \
								    ; \
   movl ADDR, %edi                                                  ; \
   movl POLYSEG_TEXTURE(%esi), %esi                                 ; \
   movl W, %ecx                                                     ; \
								    ; \
   .align 4                                                         ; \
1:                                                                  ; \
   PFRCP_R(0, 3)                /* pfrcp %mm0, %mm3 */              ; \
   PFMUL_R(1, 3)                /* pfmul %mm1, %mm3 */              ; \
   PFADD_S(DFZ_N, 0)            /* pfadd st, %mm0 */                ; \
   PF2ID_R(3, 3)                /* pf2id %mm3, %mm3 */              ; \
   PFADD_S(DFV_N, 1)            /* pfadd st, %mm1 */                ; \
								    ; \
   movq %mm3, %mm5                                                  ; \
   paddd %mm2, %mm5                                                 ; \
   psrld $1, %mm5                                                   ; \
   movq %mm2, %mm4                                                  ; \
   punpckldq %mm5, %mm2                                             ; \
   punpckhdq %mm5, %mm4                                             ; \
                                                                    ; \
   psrad VSHIFT, %mm2                                               ; \
   psrad $16, %mm4                                                  ; \
   PAND_S(VMASK_N, 2)           /* pand st, %mm2 */                 ; \
   PAND_S(UMASK_N, 4)           /* pand st, %mm4 */                 ; \
   subl $2, %ecx                                                    ; \
   paddd %mm2, %mm4                                                 ; \
   movq %mm3, %mm2                                                  ; \
   movd %mm4, %ebx                                                  ; \
   jl 77f                                                           ; \
   psrlq $32, %mm4                                                  ; \
   movd %mm4, %edx

#define END_PTEX_3D()                                                 \
   cmpl $0, %ecx                                                    ; \
   jg 1b                                                            ; \
99:                                                                 ; \
   FEMMS_R()                                                        ; \
   popl %edi                                                        ; \
   popl %esi                                                        ; \
   popl %ebx                                                        ; \
   movl %ebp, %esp                                                  ; \
   popl %ebp



/* helper for starting an fpu 1/z division */
#define START_FP_DIV()                                                \
   fld1                                                             ; \
   fdiv %st(3), %st(0)

/* helper for ending an fpu division, returning corrected u and v values */
#define END_FP_DIV()                                                  \
   fld %st(0)                    /* duplicate the 1/z value */      ; \
   fmul %st(3), %st(0)           /* divide u by z */                ; \
   fxch %st(1)                                                      ; \
   fmul %st(2), %st(0)           /* divide v by z */

#define UPDATE_FP_POS_4()                                             \
   fadds DFV                     /* update v coord */               ; \
   fxch %st(1)                   /* swap vuz stack to uvz */        ; \
   fadds DFU                     /* update u coord */               ; \
   fxch %st(2)                   /* swap uvz stack to zvu */        ; \
   fadds DFZ                     /* update z value */               ; \
   fxch %st(2)                   /* swap zvu stack to uvz */        ; \
   fxch %st(1)                   /* swap uvz stack back to vuz */

/* main body of the perspective-correct texture mapping routine
   special version for PTEX LIT with MMX */
#define DO_PTEX_LIT()                                                 \
   pushl %ebp                                                       ; \
   movl %esp, %ebp                                                  ; \
   subl $128, %esp               /* local variables */              ; \
   andl $-8, %esp                                                   ; \
   pushl %ebx                                                       ; \
   pushl %esi                                                       ; \
   pushl %edi                                                       ; \
								    ; \
   movl INFO, %esi               /* load registers */               ; \
								    ; \
   INIT()                                                           ; \
								    ; \
   flds POLYSEG_Z(%esi)          /* z at bottom of fpu stack */     ; \
   flds POLYSEG_FU(%esi)         /* followed by u */                ; \
   flds POLYSEG_FV(%esi)         /* followed by v */                ; \
   START_FP_DIV()                                                   ; \
								    ; \
   movl $0x01000000, %edx        /* add 2 to exponent */            ; \
   movl POLYSEG_DFV(%esi), %eax                                     ; \
   movl POLYSEG_DFU(%esi), %ebx                                     ; \
   movl POLYSEG_DZ(%esi), %ecx                                      ; \
   addl %edx, %eax                                                  ; \
   addl %edx, %ebx                                                  ; \
   addl %edx, %ecx                                                  ; \
   movl %eax, DFV                                                   ; \
   movl %ebx, DFU                                                   ; \
   movl %ecx, DFZ                                                   ; \
								    ; \
   movl POLYSEG_VSHIFT(%esi), %ecx                                  ; \
   movl POLYSEG_VMASK(%esi), %eax                                   ; \
   movl POLYSEG_UMASK(%esi), %edx                                   ; \
   shll %cl, %eax           /* adjust v mask and shift value */     ; \
   negl %ecx                                                        ; \
   addl $16, %ecx                                                   ; \
   movl %eax, VMASK                                                 ; \
   movl %edx, UMASK                                                 ; \
								    ; \
   END_FP_DIV()                                                     ; \
   fistpl V1                     /* store v */                      ; \
   fistpl U1                     /* store u */                      ; \
   UPDATE_FP_POS_4()                                                ; \
   START_FP_DIV()                                                   ; \
								    ; \
   movl ADDR, %edi                                                  ; \
   movl POLYSEG_TEXTURE(%esi), %esi                                 ; \
								    ; \
   movl V1, %eax                                                    ; \
   movl U1, %ebx                                                    ; \
   movl %eax, V                                                     ; \
   movl %ebx, U                                                     ; \
   sarl %cl, %eax                                                   ; \
   sarl $16, %ebx                                                   ; \
   andl VMASK, %eax                                                 ; \
   andl UMASK, %ebx                                                 ; \
   addl %eax, %ebx                                                  ; \
								    ; \
   fstps SM                                                         ; \
   fstps SV                                                         ; \
   fstps SU                                                         ; \
   fstps SZ                                                         ; \
								    ; \
   LOAD()                                                           ; \
   PIXEL1()                                                         ; \
   decl W                                                           ; \
   jle 99f                                                          ; \
                                                                    ; \
   .align 4                                                         ; \
1:                                                                  ; \
   SAVE()                                                           ; \
   emms                                                             ; \
                                                                    ; \
   flds SZ                                                          ; \
   flds SU                                                          ; \
   flds SV                                                          ; \
   flds SM                                                          ; \
								    ; \
   END_FP_DIV()                 /* finish the divide */             ; \
   fistpl V1                                                        ; \
   fistpl U1                                                        ; \
   UPDATE_FP_POS_4()            /* 4 pixels away */                 ; \
   START_FP_DIV()                                                   ; \
								    ; \
   movl V1, %eax                                                    ; \
   movl U1, %ebx                                                    ; \
   subl V, %eax                                                     ; \
   subl U, %ebx                                                     ; \
   sarl $2, %eax                                                    ; \
   sarl $2, %ebx                                                    ; \
   movl %eax, DV                                                    ; \
   movl %ebx, DU                                                    ; \
								    ; \
   addl V, %eax                                                     ; \
   addl U, %ebx                                                     ; \
   movl %eax, V                                                     ; \
   movl %ebx, %edx                                                  ; \
   sarl %cl, %eax                                                   ; \
   sarl $16, %ebx                                                   ; \
   andl VMASK, %eax                                                 ; \
   andl UMASK, %ebx                                                 ; \
   addl %eax, %ebx                                                  ; \
                                                                    ; \
   decl W                                                           ; \
   jle 98f                                                          ; \
                                                                    ; \
   movl V, %eax                                                     ; \
   addl DV, %eax                                                    ; \
   addl DU, %edx                                                    ; \
   movl %eax, V                                                     ; \
   movl %edx, U                                                     ; \
   sarl %cl, %eax                                                   ; \
   sarl $16, %edx                                                   ; \
   andl VMASK, %eax                                                 ; \
   andl UMASK, %edx                                                 ; \
   addl %eax, %edx                                                  ; \
                                                                    ; \
   fstps SM                                                         ; \
   fstps SV                                                         ; \
   fstps SU                                                         ; \
   fstps SZ                                                         ; \
								    ; \
   LOAD()                                                           ; \
   PIXEL2()                                                         ; \
   decl W                                                           ; \
   jle 99f                                                          ; \
                                                                    ; \
   movl V, %eax                                                     ; \
   movl U, %ebx                                                     ; \
   addl DV, %eax                                                    ; \
   addl DU, %ebx                                                    ; \
   sarl %cl, %eax                                                   ; \
   sarl $16, %ebx                                                   ; \
   andl VMASK, %eax                                                 ; \
   andl UMASK, %ebx                                                 ; \
   addl %eax, %ebx                                                  ; \
                                                                    ; \
   decl W                                                           ; \
   jle 97f                                                          ; \
                                                                    ; \
   movl V1, %eax                                                    ; \
   movl U1, %edx                                                    ; \
   movl %eax, V                                                     ; \
   movl %edx, U                                                     ; \
   sarl %cl, %eax                                                   ; \
   sarl $16, %edx                                                   ; \
   andl VMASK, %eax                                                 ; \
   andl UMASK, %edx                                                 ; \
   addl %eax, %edx                                                  ; \
                                                                    ; \
   PIXEL2()                                                         ; \
   decl W                                                           ; \
   jg 1b                                                            ; \
   nop                                                              ; \
   jmp 99f                                                          ; \
                                                                    ; \
   .align 4                                                         ; \
98:                                                                 ; \
   fstps SM                                                         ; \
   fstps SV                                                         ; \
   fstps SU                                                         ; \
   fstps SZ                                                         ; \
								    ; \
   LOAD()                                                           ; \
97:                                                                 ; \
   PIXEL1()                                                         ; \
99:                                                                 ; \
   emms                                                             ; \
   popl %edi                                                        ; \
   popl %esi                                                        ; \
   popl %ebx                                                        ; \
   movl %ebp, %esp                                                  ; \
   popl %ebp



#ifdef ALLEGRO_COLOR16
.globl __poly_scanline_ptex_lit15x
   .align 4
__poly_scanline_ptex_lit15x:

#define INIT()                                                          \
   INIT_HIGH(15)                                                      ; \
   movq %mm7, ALPHA                                                   ; \
   emms

#define LOAD()                                                          \
   movq ALPHA, %mm7

#define PIXEL1()                                                        \
   movw (%esi, %ebx, 2), %ax                                          ; \
   ONE_HIGH(15)                                                       ; \
   paddd DALPHA, %mm7                                                 ; \
   addl $2, %edi

#define PIXEL2()                                                        \
   TWO_HIGH(15)

#define SAVE()                                                          \
   movq %mm7, ALPHA

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_lit15x() */

#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_mask_lit15x
   .align 4
__poly_scanline_ptex_mask_lit15x:

#define PIXEL1()                                                        \
   MASK_HIGH(15)

#define PIXEL2()                                                        \
   MASK_HIGH(15)                                                      ; \
   movl %edx, %ebx                                                    ; \
   MASK_HIGH(15)

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_mask_lit15x() */

#undef INIT
#undef LOAD
#undef SAVE
#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_lit15d
   .align 4
__poly_scanline_ptex_lit15d:
   INIT_PTEX_3D(
      INIT_HIGH(15)
   )
   TWO_HIGH(15)
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_lit15d() */
   .align 4
77:
   movw (%esi, %ebx, 2), %ax
   ONE_HIGH(15)
   jmp 99b

.globl __poly_scanline_ptex_mask_lit15d
   .align 4
__poly_scanline_ptex_mask_lit15d:
   INIT_PTEX_3D(
      INIT_HIGH(15)
   )
   MASK_HIGH(15)
   movl %edx, %ebx
77:
   MASK_HIGH(15)
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_mask_lit15d() */



.globl __poly_scanline_ptex_lit16x
   .align 4
__poly_scanline_ptex_lit16x:

#define INIT()                                                          \
   INIT_HIGH(16)                                                      ; \
   movq %mm7, ALPHA                                                   ; \
   emms

#define LOAD()                                                          \
   movq ALPHA, %mm7

#define PIXEL1()                                                        \
   movw (%esi, %ebx, 2), %ax                                          ; \
   ONE_HIGH(16)                                                       ; \
   paddd DALPHA, %mm7                                                 ; \
   addl $2, %edi

#define PIXEL2()                                                        \
   TWO_HIGH(16)

#define SAVE()                                                          \
   movq %mm7, ALPHA

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_lit16x() */

#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_mask_lit16x
   .align 4
__poly_scanline_ptex_mask_lit16x:

#define PIXEL1()                                                        \
   MASK_HIGH(16)

#define PIXEL2()                                                        \
   MASK_HIGH(16)                                                      ; \
   movl %edx, %ebx                                                    ; \
   MASK_HIGH(16)

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_mask_lit16x() */

#undef INIT
#undef LOAD
#undef SAVE
#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_lit16d
   .align 4
__poly_scanline_ptex_lit16d:
   INIT_PTEX_3D(
      INIT_HIGH(16)
   )
   TWO_HIGH(16)
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_lit16d() */
   .align 4
77:
   movw (%esi, %ebx, 2), %ax
   ONE_HIGH(16)
   jmp 99b

.globl __poly_scanline_ptex_mask_lit16d
   .align 4
__poly_scanline_ptex_mask_lit16d:
   INIT_PTEX_3D(
      INIT_HIGH(16)
   )
   MASK_HIGH(16)
   movl %edx, %ebx
77:
   MASK_HIGH(16)
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_mask_lit16d() */
#endif /* COLOR16 */



#ifdef ALLEGRO_COLOR32
.globl __poly_scanline_ptex_lit32x
   .align 4
__poly_scanline_ptex_lit32x:

#define INIT()                                                          \
   INIT_TRUE(32)                                                      ; \
   movq %mm7, ALPHA                                                   ; \
   emms

#define LOAD()                                                          \
   movq ALPHA, %mm7

#define PIXEL1()                                                        \
   ONE_32()                                                           ; \
   paddd DALPHA, %mm7                                                 ; \
   addl $4, %edi

#define PIXEL2()                                                        \
   TWO_32()

#define SAVE()                                                          \
   movq %mm7, ALPHA

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_lit32x() */

#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_mask_lit32x
   .align 4
__poly_scanline_ptex_mask_lit32x:

#define PIXEL1()                                                        \
   MASK_32()

#define PIXEL2()                                                        \
   MASK_32()                                                          ; \
   movl %edx, %ebx                                                    ; \
   MASK_32()

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_mask_lit32x() */

#undef INIT
#undef LOAD
#undef SAVE
#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_lit32d
   .align 4
__poly_scanline_ptex_lit32d:
   INIT_PTEX_3D(
      INIT_TRUE(32)
   )
   TWO_32()
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_lit32d() */
   .align 4
77:
   ONE_32()
   jmp 99b

.globl __poly_scanline_ptex_mask_lit32d
   .align 4
__poly_scanline_ptex_mask_lit32d:
   INIT_PTEX_3D(
      INIT_TRUE(32)
   )
   MASK_32()
   movl %edx, %ebx
77:
   MASK_32()
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_mask_lit32d() */
#endif /* COLOR32 */



#ifdef ALLEGRO_COLOR24
.globl __poly_scanline_ptex_lit24x
   .align 4
__poly_scanline_ptex_lit24x:

#define INIT()                                                          \
   INIT_TRUE(24)                                                      ; \
   movq %mm7, ALPHA                                                   ; \
   emms

#define LOAD()                                                          \
   movq ALPHA, %mm7

#define PIXEL1()                                                        \
   ONE_24()                                                           ; \
   paddd DALPHA, %mm7                                                 ; \
   addl $3, %edi

#define PIXEL2()                                                        \
   TWO_24()

#define SAVE()                                                          \
   movq %mm7, ALPHA

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_lit24x() */

#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_mask_lit24x
   .align 4
__poly_scanline_ptex_mask_lit24x:

#define PIXEL1()                                                        \
   MASK_24()

#define PIXEL2()                                                        \
   MASK_24()                                                          ; \
   movl %edx, %ebx                                                    ; \
   MASK_24()

   DO_PTEX_LIT()
   ret                           /* end of _poly_scanline_ptex_mask_lit24x() */

#undef INIT
#undef LOAD
#undef SAVE
#undef PIXEL1
#undef PIXEL2

.globl __poly_scanline_ptex_lit24d
   .align 4
__poly_scanline_ptex_lit24d:
   INIT_PTEX_3D(
      INIT_TRUE(24)
   )
   TWO_24()
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_lit24d() */
   .align 4
77:
   ONE_24()
   jmp 99b

.globl __poly_scanline_ptex_mask_lit24d
   .align 4
__poly_scanline_ptex_mask_lit24d:
   INIT_PTEX_3D(
      INIT_TRUE(24)
   )
   MASK_24()
   movl %edx, %ebx
77:
   MASK_24()
   END_PTEX_3D()
   ret                           /* end of _poly_scanline_ptex_mask_lit24d() */
#endif /* COLOR24 */

#endif /* MMX */
